;;; osm.lisp
;;;
;;; Author: Eric Marsden <eric.marsden@free.fr>
;;; Time-stamp: <2007-12-29 emarsden> 
;;;
;;
;; === TODO ===
;;
;; - add Landsat imagery for the background
;; - render way names
;; - modulate way rendering according to tags


;; Landsat etc: http://worldkit.org/doc/layers.php


(in-package :osm)


(defclass osm-object ()
  ((id          :initarg :id
                :accessor id
                :type integer)
   (timestamp   :initarg :timestamp
                :initform -1
                :accessor timestamp
                :type fixnum)
   (tags        :initarg :tags
                :initform nil
                :accessor tags)))

(defclass node (osm-object)
  ((lat   :initarg :lat
          :accessor latitude)
   (lon   :initarg :lon
          :accessor longitude)))

(defclass way (osm-object)
  ((nodes   :initarg :nodes
            :accessor nodes)))

(defmethod name ((way way))
  (with-slots (tags) way
    (or (cdr (assoc "name" tags :test #'string=))
        (cdr (assoc "ref" tags :test #'string=))
        (cdr (assoc "int_ref" tags :test #'string=)))))


(defmethod print-object ((object node) (stream stream))
  (with-slots (id lat lon) object
    (print-unreadable-object (object stream :type t :identity nil)
      (format stream "#~d @ ~f,~f" id lat lon))))

(defmethod print-object ((object way) (stream stream))
  (with-slots (id) object
    (print-unreadable-object (object stream :type t :identity nil)
      (format stream "#~d~@[ ~S~]" id (name object)))))



(defvar *nodes* (make-hash-table :test #'eql))

(defvar *ways* (make-hash-table :test #'eql))

(defun reset-database ()
  (setq *nodes* (make-hash-table :test #'eql)       
        *ways* (make-hash-table :test #'eql)))

(defun make-node (id lat lon &key timestamp tags)
  (let ((node (make-instance 'node :id id
                             :lat lat
                             :lon lon
                             :timestamp timestamp
                             :tags tags)))
    (setf (gethash id *nodes*) node)
    node))

(defun make-way (id nodes &key timestamp tags)
  (let ((way (make-instance 'way :id id
                            :nodes nodes
                            :timestamp timestamp
                            :tags tags)))
    (setf (gethash id *ways*) way)
    way))

(defun find-node (id)
  (or (gethash id *nodes*)
      (error "No node with id ~A" id)))

(defun find-way (id)
  (or (gethash id *ways*)
      (error "No way with id ~A" id)))

(defun map-nodes (thunk)
  (maphash (lambda (key value) (funcall thunk value)) *nodes*))

(defun map-ways (thunk)
  (maphash (lambda (key value) (funcall thunk value)) *ways*))


(defmethod add-node ((node node) (way way))
  (push node (nodes way)))

(defmethod add-tag ((object osm-object) key value)
  (push (cons key value) (tags object)))



;; EOF
