;;; parse-osm.lisp
;;;
;;; Author: Eric Marsden <eric.marsden@free.fr>
;;; Time-stamp: <2007-12-29 emarsden>
;;;
;;
;; Parse the contents of an OSM file (in XML format). Can be used to
;; parse a planet.osm file, a subset created with JOSM or osmosis, or
;; the output from the REST API.


(in-package :osm)

(defun format-iso8601-time (time-value &optional include-timezone-p)
    "Formats a universal time TIME-VALUE in ISO 8601 format, with the time zone included if INCLUDE-TIMEZONE-P is non-NIL"
    (flet ((format-iso8601-timezone (zone)
             (if (zerop zone)
               "Z"
               (multiple-value-bind (h m) (truncate (abs zone) 1.0)
                 ;; Tricky.  Sign of time zone is reversed in ISO 8601
                 ;; relative to Common Lisp convention!
                 (format nil "~:[+~;-~]~2,'0D:~2,'0D"
                         (> zone 0) h (round m))))))
    (multiple-value-bind (second minute hour day month year dow dst zone)
      (decode-universal-time time-value)
      (declare (ignore dow dst))
      (format nil "~4,'0D-~2,'0D-~2,'0DT~2,'0D:~2,'0D:~2,'0D~:[~*~;~A~]"
              year month day hour minute second
              include-timezone-p (format-iso8601-timezone zone)))))

(defun parse-iso8601-time (time-string)
  "Parses an ISO 8601 format string and returns the universal time"
  (flet ((parse-delimited-string (string delimiter n)
           ;; Parses a delimited string and returns a list of
           ;; n integers found in that string.
           (let ((answer (make-list n :initial-element 0)))
             (loop for i upfrom 0
                   for start = 0 then (1+ end)
                   for end = (position delimiter string :start (1+ start))
                   do (setf (nth i answer)
                            (parse-integer (subseq string start end)))
                   when (null end) return t)
             (values-list answer)))
         (parse-fixed-field-string (string field-sizes)
           ;; Parses a string with fixed length fields and returns
           ;; a list of integers found in that string.
           (let ((answer (make-list (length field-sizes) :initial-element 0)))
             (loop with len = (length string)
                   for start = 0 then (+ start field-size)
                   for field-size in field-sizes
                   for i upfrom 0
                   while (< start len)
                   do (setf (nth i answer)
                            (parse-integer (subseq string start (+ start field-size)))))
             (values-list answer))))
    (flet ((parse-iso8601-date (date-string)
             (let ((hyphen-pos (position #\- date-string)))
               (if hyphen-pos
                 (parse-delimited-string date-string #\- 3)
                 (parse-fixed-field-string date-string '(4 2 2)))))
           (parse-iso8601-timeonly (time-string)
             (let* ((colon-pos (position #\: time-string))
                    (zone-pos (or (position #\- time-string)
                                  (position #\+ time-string)
                                  (position #\Z time-string)))
                    (timeonly-string (subseq time-string 0 zone-pos))
                    (zone-string (when zone-pos (subseq time-string (1+ zone-pos))))
                    (time-zone nil))
               (when zone-pos
                 (unless (zerop (length zone-string))
                   (multiple-value-bind (zone-h zone-m)
                       (parse-delimited-string zone-string #\: 2)
                     (setq time-zone (+ zone-h (/ zone-m 60)))
                     (when (char= (char time-string zone-pos) #\-)
                       (setq time-zone (- time-zone))))))
               (multiple-value-bind (hh mm ss)
                   (if colon-pos
                       (parse-delimited-string timeonly-string #\: 3)
                       (parse-fixed-field-string timeonly-string '(2 2 2)))
                 (values hh mm ss time-zone)))))
    (let ((time-separator (position #\T time-string)))
      (multiple-value-bind (year month date)
                           (parse-iso8601-date
                            (subseq time-string 0 time-separator))
        (if time-separator
          (multiple-value-bind (hh mm ss zone)
                               (parse-iso8601-timeonly
                                (subseq time-string (1+ time-separator)))
            (if zone
              ;; Tricky:  Sign of time zone is reversed in ISO 8601
              ;; relative to Common Lisp convention!
              (encode-universal-time ss mm hh date month year (- zone))
              (encode-universal-time ss mm hh date month year)))
          (encode-universal-time 0 0 0 date month year)))))))

;; (parse-iso8601-time "2007-10-08T17:03:50+01:00") => 3400848230
;; (parse-iso8601-time "2007-10-08T17:03:50Z") => 3400844630


(defun parse-timestamp (string)
  (parse-iso8601-time string))


(defun parse-osm-xml (xml-octets)
  (klacks:with-open-source (s (cxml:make-source xml-octets))
    (loop
     :for key = (klacks:peek s)
     :with current-object = nil
     :with *read-eval* = nil
     :while key :do
       (case key
         (:start-element
          (let ((qname (klacks:current-qname s)))
            ;; <node id="648212" lat="43.5752045929005" lon="1.43638183139437" visible="true" timestamp="2006-07-05T17:47:40+01:00"/>
            (cond ((string= "node" qname)
                   (setq current-object
                         (make-node (parse-integer (klacks:get-attribute s "id"))
                                    (read-from-string (klacks:get-attribute s "lat"))
                                    (read-from-string (klacks:get-attribute s "lon"))
                          :timestamp (parse-timestamp (klacks:get-attribute s "timestamp")))))
                  
                  ((string= "way" qname)
                   (setq current-object (make-way (parse-integer (klacks:get-attribute s "id")) nil
                                         :timestamp (parse-timestamp (klacks:get-attribute s "timestamp")))))
                  
                  ((string= "nd" qname)
                   (unless (typep current-object 'way)
                     (error "nd tag not within a way"))
                   (unless (eql 1 (length (klacks:list-attributes s)))
                     (error "nd tag contains spurious tags ~A" (klacks:list-attributes s)))
                   (let* ((node-id (parse-integer (klacks:get-attribute s "ref")))
                          (node (find-node node-id)))
                     (add-node node current-object)))
                   
                  ((string= "tag" qname)
                   (add-tag current-object (klacks:get-attribute s "k") (klacks:get-attribute s "v")))))))
       (klacks:consume s))))



(defun slurp-osm-file (filespec)
  (with-open-file (in filespec :direction :input
                      :element-type '(unsigned-byte 8))
    (parse-osm-xml in)))
       


;; EOF
