;;; render-history.lisp -- rendering the time of data insertion
;;;
;;; Author: Eric Marsden <eric.marsden@free.fr>
;;; Time-stamp: <2007-12-30 emarsden>
;;
;;
;;
;; France OSM: http://hexagone.openstreetmap.fr/
;;
;; Similar for GPX tracks in Python:
;; http://svn.openstreetmap.org/applications/rendering/party/video.py


(in-package :osm)


;; displays a text histogram of the age of elements inserted into the OSM database
(defun print-histogram (&optional (bucket-count 100))
  (let ((oldest most-positive-fixnum)
        (youngest -1)
        (buckets (make-array (1+ bucket-count) :element-type 'fixnum :initial-element 0))
        (bucket-size nil)
        (max-bucket-fill 0))
    (map-nodes
     (lambda (n)
       (setf oldest (min (timestamp n) oldest))
       (setf youngest (max (timestamp n) youngest))))
    (map-ways
     (lambda (way)
       (setf oldest (min (timestamp way) oldest))
       (setf youngest (max (timestamp way) youngest))))
    (format *debug-io* "Got elements from ~A to ~A~%"
            (format-iso8601-time oldest)
            (format-iso8601-time youngest))
    
    ;; now fill up the buckets
    (setq bucket-size (floor (- youngest oldest) bucket-count))
    (flet ((bucketize (time)
             (floor (- time oldest) bucket-size)))
      (map-nodes (lambda (n) (incf (aref buckets (bucketize (timestamp n))))))
      (map-ways (lambda (way) (incf (aref buckets (bucketize (timestamp way)))))))
    (dotimes (bucket bucket-count)
      (setf max-bucket-fill (max max-bucket-fill (aref buckets bucket))))
    (dotimes (bucket bucket-count)
      (multiple-value-bind (second minute hour day month year)
          (decode-universal-time (+ oldest (* bucket bucket-size)))
        (declare (ignore second minute hour day))
        (format t "~4,'0D-~2,'0D: " year month))
      (dotimes (i (round (* 80 (/ (aref buckets bucket) max-bucket-fill))))
        (princ "*"))
      (terpri))))


;; R, G, B are floating point in [0, 1]
(defun rgb/skippy (r g b)
  (flet ((hexify (f) (round (* f 255))))
    (skippy:rgb-color (hexify r) (hexify g) (hexify b))))

(defvar *rgb-consumer* #'rgb/skippy)

(defun greyscale-color (ratio)
  (funcall *rgb-consumer*  ratio ratio ratio))

(defun hsv-color (hue &optional (saturation 0.99) (value 0.99))
  (multiple-value-bind (sextant f)
      (floor (* 6.0 hue))
    (let ((p (* value (- 1 saturation)))
          (q (* value (- 1 (* saturation f))))
          (r (* value (- 1 (* saturation (- 1 f))))))
      (case sextant
        (0  (funcall *rgb-consumer* value r p))
        (1  (funcall *rgb-consumer* q value p))
        (2  (funcall *rgb-consumer* p value r))
        (3  (funcall *rgb-consumer* p q value))
        (4  (funcall *rgb-consumer* r p value))
        (5  (funcall *rgb-consumer* value p q))))))

(defun sin-color (ratio)
  (let ((angle (* ratio pi)))
    (apply *rgb-consumer*
             (mapcar (lambda (angle)
                       (let ((s (/ (+ 1 (sin angle)) 2))) (* s s)))
                     (list angle
                           (+ angle (/ (* pi 4.0) 3.0))
                           (+ angle (/ (* pi 3.0) 3.0)))))))


;; generate an animated GIF showing the historical buildup of node data
(defun GIF-node-history (&key (width 200) (height 200) (frames 100))
  (declare (optimize (debug 3) (speed 0)))
  (let* ((oldest most-positive-fixnum)
         (youngest -1)
         (frame-length nil)
         (color-table (skippy:make-color-table))
         (white (skippy:ensure-color #xFFFFFF color-table))
         (data-stream (skippy:make-data-stream :color-table color-table
                                               :loopingp t
                                               :height height
                                               :width width))
         (x-min 1e6)
         (y-min 1e6)
         (x-max -1e6)
         (y-max -1e6)
         (x-range 0)
         (y-range 0))
    ;; initialize the color-table
    (dotimes (i 254)
      (skippy:ensure-color (hsv-color (/ i 255.0)) color-table))
    (map-nodes
     (lambda (n)
       (setf oldest (min (timestamp n) oldest))
       (setf youngest (max (timestamp n) youngest))
       (with-slots (lat lon) n
         (when (< lon x-min) (setf x-min lon))
         (when (> lon x-max) (setf x-max lon))
         (when (< lat y-min) (setf y-min lat))
         (when (> lat y-max) (setf y-max lat)))))
    (setq x-range (- x-max x-min)
          y-range (- y-max y-min))    
    (format *debug-io* "Got elements from ~A to ~A~%"
            (format-iso8601-time oldest)
            (format-iso8601-time youngest))
    (setq frame-length (floor (- youngest oldest) frames))
    (labels ((unscale-x (x)
               (floor (* (1- width) (/ (- x x-min) x-range))))
             (unscale-y (y)
               (floor (* (1- height) (/ (- y y-min) y-range)))))
      (dotimes (frame frames)
        (let* ((cutoff-date/lower (+ oldest (* frame frame-length)))
               (cutoff-date/upper (+ cutoff-date/lower frame-length))
               (skippy:*default-delay-time* 2)
               (canvas (skippy:make-canvas :height height
                                           :width width)))
          (map-nodes
           (lambda (n)
             (when (< (timestamp n) cutoff-date/upper)
               (setf (skippy:pixel-ref canvas (unscale-x (longitude n)) (unscale-y (latitude n)))
                     (1+ (floor (* 254.0 (/ (- (timestamp n) oldest) (- youngest oldest)))))))))
          (setf (skippy:disposal-method (skippy:canvas-image canvas)) :none)
          (setf (skippy:transparency-index (skippy:canvas-image canvas)) 0)
          (skippy:add-image (skippy:canvas-image canvas) data-stream)
          #+nil (skippy:add-delay 2 data-stream))))
    (skippy:output-data-stream data-stream #p"/tmp/GIF-node-history.gif")))
      


;; create a series of PNGs showing epochs at which nodes and ways were
;; added to the map. The PNGs can then be animated in a movie using
;; mencoder or ffmpeg.
;;
;; mencoder "mf://*.jpg" -mf fps=25 -o output.avi -ovc divx4
;;
;; mencoder mf://*.jpg -mf w=800:h=600:fps=25:type=jpeg -ovc lavc -lavcopts vcodec=mpeg4 -oac copy -o output.avi
;;
;; mencoder "mf://*png" -mf w=800:h=600:fps=5:type=png -ovc lavc -lavcopts vcodec=mpeg4:mbd=2:trell -oac copy -o output.avi
;;
;; ffmpeg -r 10 -i osm-history-%04d.png output.mp4
;; 
;;
;; http://www.mplayerhq.hu/DOCS/HTML/en/menc-feat-enc-images.html
(defun generate-historical-PNGs (&key (width 800) (height 600) (frames 100))
  (let ((x-min 1e6)
        (y-min 1e6)
        (x-max -1e6)
        (y-max -1e6)
        (x-range 0)
        (y-range 0)
        (oldest most-positive-fixnum)
        (youngest -1)
        (frame-length nil))
    (map-nodes
     (lambda (n)
       (with-slots (lat lon) n
         (when (< lon x-min) (setf x-min lon))
         (when (> lon x-max) (setf x-max lon))
         (when (< lat y-min) (setf y-min lat))
         (when (> lat y-max) (setf y-max lat)))))
    (setq x-range (- x-max x-min)
          y-range (- y-max y-min))
    (map-ways
     (lambda (w)
       (setf oldest (min (timestamp w) oldest))
       (setf youngest (max (timestamp w) youngest))))
    (setq frame-length (floor (- youngest oldest) frames))
    (dotimes (frame frames)
      (with-canvas (:width width :height height)
        (set-rgb-fill 1 1 1)
        (clear-canvas)
        (set-line-width 1)
        (scale (/ width x-range) (/ height y-range))
        (translate (- x-min)  (- y-min))
        (set-line-width (/ x-range 500.0))
        (let* ((cutoff-date/lower (+ oldest (* frame frame-length)))
               (cutoff-date/upper (+ cutoff-date/lower frame-length))
               (date/str (multiple-value-bind (second minute hour day month year)
                             (decode-universal-time cutoff-date/lower)
                           (declare (ignore second minute hour day))
                           (format nil "~4,'0D-~2,'0D" year month))))
          (set-font (get-font "/usr/share/fonts/truetype/ttf-dejavu/DejaVuSans.ttf") (/ x-range 60))
          (set-rgb-fill 0 0 0)
          (draw-string (+ x-min (* 0.03 x-range)) (+ y-min (* 0.03 y-range)) date/str)
          (stroke)
          #+draw-nodes
          (map-nodes
           (lambda (n)
             (when (< (timestamp n) cutoff-date/upper)
               (let ((grey (coerce (/ (- (timestamp n) oldest) (- youngest oldest)) 'float)))
                 (with-slots (lat lon) n
                   ;; (set-rgb-stroke grey grey grey)
                   (let ((*rgb-consumer* #'list))
                     (apply #'set-rgb-stroke (sin-color grey)))
                   (centered-circle-path lon lat (/ x-range 1200.0))
                   (stroke))))))
          (map-ways
           (lambda (w)
             (when (< (timestamp w) cutoff-date/upper)
               (let ((grey (coerce (/ (- (timestamp w) oldest) (- youngest oldest)) 'float))
                     (nodes (nodes w)))
                 (let ((*rgb-consumer* #'list))
                   (apply #'set-rgb-stroke (sin-color grey)))
                 (with-slots (lat lon) (first nodes)
                   (move-to lon lat))
                 (dolist (n (rest nodes))
                   (with-slots (lat lon) n
                     (line-to lon lat)))
                 (stroke))))))
        (save-png (format nil "/tmp/osm-history-~4,'0D.png" frame))))))

;; can also run optipng /tmp/osm-history*png to improve compression of
;; the PNG images, losslessly.


;; EOF
